package com.vernaio.example;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.supercsv.io.CsvListReader;
import org.supercsv.prefs.CsvPreference;

import com.fasterxml.jackson.core.JsonGenerationException;
import com.fasterxml.jackson.core.JsonParseException;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.datatype.jdk8.Jdk8Module;
import com.vernaio.aivis.engine.v2.base.sdk.dto.DtoFloatCell;
import com.vernaio.aivis.engine.v2.base.sdk.dto.DtoStringCell;
import com.vernaio.aivis.engine.v2.base.sdk.exception.EngineException;
import com.vernaio.aivis.engine.v2.base.sdk.exception.MissingLibraryException;
import com.vernaio.aivis.engine.v2.base.sdk.exception.WrongFlavourException;
import com.vernaio.aivis.engine.v2.response.analysis.sdk.ResponseAnalysisData;

public class Utils {
	
	private static final ObjectMapper MAPPER = new ObjectMapper().registerModule(new Jdk8Module());
	
	public static class FloatColumn {
		
		private final String columnId;
		
		private final List<DtoFloatCell> cells;
		
		protected FloatColumn(final String columnId) {
			this.columnId = columnId;
			this.cells = new ArrayList<DtoFloatCell>();
		}
		
		public List<DtoFloatCell> getCells() {
			return this.cells;
		}
		
		public String getColumnId() {
			return columnId;
		}
	}
	
	public static class StringColumn {
		
		private final String columnId;
		
		private final List<DtoStringCell> cells;
		
		protected StringColumn(final String columnId) {
			this.columnId = columnId;
			this.cells = new ArrayList<DtoStringCell>();
		}
		
		public List<DtoStringCell> getCells() {
			return this.cells;
		}
		
		public String getColumnId() {
			return columnId;
		}
	}
	
	/*
	 * The following method knows very well how the data csv in this example looks like, reads it and at the end constructs the data context:
	 * Columns are added to the data context by naming their column id and a list of their cells
	 * (DtoFloatCell for float columns, DtoStringCell for string columns, would be analogously for boolean columns).
	 * The method can be adapted if the data comes in a different format, e.g. via parquet files.
	 */
	public static ResponseAnalysisData readCsvConvertToAivisFormat(final File file, final List<String> stringColumnIds) throws FileNotFoundException, IOException,
		WrongFlavourException, MissingLibraryException, EngineException {
		final HashMap<String, FloatColumn> floatColumns = new HashMap<String, FloatColumn>();
		final HashMap<String, StringColumn> stringColumns = new HashMap<String, StringColumn>();
		final CsvPreference csvPreference = new CsvPreference.Builder('"', ',', "\n").ignoreEmptyLines(false).build();
		
		System.out.println(String.format("\t... reading file '%s' ...", file.getPath()));
		
		// get line numbers to show nice progress bar
		int lineCount = 0;
		try (LineNumberReader count = new LineNumberReader(new FileReader(file))) {
			count.skip(Long.MAX_VALUE);
			lineCount = count.getLineNumber();
		}
		
		// read file and construct lists of data points
		try (
			final InputStream stream = new FileInputStream(file);
			final InputStreamReader reader = new InputStreamReader(stream, StandardCharsets.UTF_8);
			final CsvListReader csvReader = new CsvListReader(reader, csvPreference)
		) {
			
			// the columns ids are in the csv's header
			final String[] header = csvReader.getHeader(true);
			for (int i = 1; i < header.length; i++) {
				final String columnId = header[i];
				if (stringColumnIds.contains(columnId)) {
					stringColumns.put(columnId, new StringColumn(columnId));
				} else {
					floatColumns.put(columnId, new FloatColumn(columnId));
				}
			}
			
			System.out.print("\t... processing columns ...");
			
			// show nice progress bar
			int progressBar = -1;
			int line = 0;
			
			// read values by index
			List<String> values;
			while ((values = csvReader.read()) != null) {
				
				// show nice progress bar
				final int progressBarNew = line / (lineCount / 10) * 10;
				if (progressBar != progressBarNew) {
					progressBar = progressBarNew;
					// System.out.print("\r");
					System.out.print(String.format("\r\t... processing columns: %d%% ...", progressBar));
				}
				
				// to this row, add cells to all columns (if value exists)
				for (int i = 1; i < values.size(); i++) {
					final String stringValue = values.get(i);
					if (stringValue != null && !stringValue.isEmpty()) {
						
						final String columnId = header[i];
						// we know: the 0th column is the column containing the row id
						if (stringColumnIds.contains(columnId)) {
							stringColumns.get(columnId).getCells().add(new DtoStringCell(String.valueOf(values.get(0)), stringValue));
						} else {
							final double doubleValue = Double.parseDouble(stringValue);
							floatColumns.get(columnId).getCells().add(new DtoFloatCell(String.valueOf(values.get(0)), doubleValue));
						}
					}
				}
				
				line++;
			}
		}
		
		System.out.print(String.format("\r\t... processing columns: %d%% ...", 100));
		
		// finally create data context
		System.out.println("and now adding columns to data context:");
		final ResponseAnalysisData responseAnalysisData = ResponseAnalysisData.create();
		
		for (Map.Entry<String, FloatColumn> column : floatColumns.entrySet()) {
			responseAnalysisData.addFloatColumn(column.getKey(), ((FloatColumn) column.getValue()).getCells());
			System.out.println(String.format("\t... float column '%s' has %s values", column.getKey(), ((FloatColumn) column.getValue()).getCells().size()));
		}
		for (Map.Entry<String, StringColumn> column : stringColumns.entrySet()) {
			responseAnalysisData.addStringColumn(column.getKey(), ((StringColumn) column.getValue()).getCells());
			System.out.println(String.format("\t... string column '%s' has %s values", column.getKey(), ((StringColumn) column.getValue()).getCells().size()));
		}
		System.out.println("\t... data context is finished!");
		
		return responseAnalysisData;
	}
	
	public static <T> T readJson(final File file, final Class<T> clazz) throws JsonParseException, JsonMappingException, IOException {
		System.out.println(String.format("\t... reading file '%s' ...", file.getPath()));
		return MAPPER.readValue(file, clazz);
	}
	
	public static <T> void writeJson(final File file, final T obj) throws JsonGenerationException, JsonMappingException, IOException {
		System.out.println(String.format("\t... writing file '%s' ...", file.getPath()));
		Optional.ofNullable(file.getParentFile()).ifPresent(f -> f.mkdirs());
		MAPPER.writeValue(file, obj);
	}
	
}
