# this python script performs the same calculation as example_ra.ipynb

import json
import csv
import argparse
import os

from aivis_engine_v2_ra_sdk_python import (
    DtoFloatCell,
    DtoStringCell,
    ResponseAnalysisSetup,
    ResponseAnalysisData,
    ResponseAnalysis,
    EngineLogger,
)

# ---------------------------------------------------------------------------------
# Data location

INPUT_DATA_FILE = "train_ra.csv"
TARGET_COLUMN_ID = "TARGET"


# Input and output folder specified via args
def from_args():
    parser = argparse.ArgumentParser(
        description="Example project for response analysis"
    )
    # data is expected in input folder
    parser.add_argument(
        "--input", "-i", type=str, nargs=1, required=True, help="Path to input folder"
    )
    # output files will be written to output folder
    parser.add_argument(
        "--output", "-o", type=str, nargs=1, required=True, help="Path to output folder"
    )
    args = parser.parse_args()
    # create output folder if not existing
    if not os.path.exists(args.output[0]):
        os.mkdir(args.output[0])
    return {
        "input_folder": args.input[0],
        "output_folder": args.output[0],
    }


# ---------------------------------------------------------------------------------
# Logger and Threads


class Logger(EngineLogger):
    def log(self, level, thread, module, message):
        min_logging_level = 3
        if level <= min_logging_level:
            print("\t... %s" % message)


def setup_logger_and_threads():
    print("Initializing ResponseAnalysisSetup: logging and threads ...")
    ResponseAnalysisSetup.register_logger(Logger())
    max_threads = 0  # number of threads will be chosen automatically with 0
    ResponseAnalysisSetup.init_thread_count(max_threads)
    print("\t... done!")
    return


# ---------------------------------------------------------------------------------
# Reading and Writing


def read_json(filename):
    print("\t... reading file '%s' ..." % filename)
    with open(filename, "r") as file:
        return json.loads(file.read())


def write_json(filename, obj):
    print("\t... writing file '%s' ..." % filename)
    with open(filename, "w") as file:
        file.write(json.dumps(obj, indent=2))
    return


CHUNK_SIZE = 1024


# Main data reading function
# Float columns are added to an aivis data context via their id and a list of DtoFloatCell.
# (Analogously for boolean and string columns.)
# The rest of the following function can be adapted if the data comes in a different format as in this example, e.g. in parquet files.
def read_csv_convert_to_aivis_format(filename):
    print("\t... reading file '%s' ..." % filename)

    # create empty data context for input data
    data_context = ResponseAnalysisData.create()

    # get line numbers to show nice progress bar
    line_count = sum(1 for _ in open(filename))

    # list of string columns
    string_columns = [
        "status_of_existing_checking_account",
        "purpose",
        "savings_account",
        "credit_history",
        "marital_status_and_sex",
        "other_debtors",
        "property",
        "other_installment_plans",
        "housing",
        "job",
        "telephone",
        "foreign_worker",
    ]

    # Read data in chunks to keep total memory usage low
    def consume_cells(consume_columns):
        for column in consume_columns:
            if column["column_id"] in string_columns:
                data_context.add_string_column(
                    column["column_id"], column["cells"]
                )
            else:
                data_context.add_float_column(column["column_id"], column["cells"])
            column["cells"].clear()

    # read file
    with open(filename, newline="\n") as csvfile:
        # in this example, column ids are in the header, first column is id
        reader = csv.reader(csvfile, delimiter=",", quotechar='"')
        headers = next(reader)
        columns = list(map(lambda h: {"column_id": h, "cells": []}, headers[1:]))

        # show nice progress bar
        progress_bar = -1
        line = 0

        nchunks = 0

        for n, row in enumerate(reader, start=1):
            # show nice progress bar
            progess_bar_new = round(line / (line_count / 10)) * 10
            if progress_bar != progess_bar_new:
                progress_bar = progess_bar_new
                print("\r\t... processing columns: %d%% ..." % progress_bar, end="")
            line += 1

            # get row_id of this row, which is in the csv's first column
            row_id = int(row[0])

            # add cells to all column's list (if value exists)
            for i, column in enumerate(row[1:], start=0):
                if column:
                    if columns[i]["column_id"] in string_columns:
                        dtosdp = DtoStringCell(str(row_id), column)
                        columns[i]["cells"].append(dtosdp)
                    else:
                        dtofdp = DtoFloatCell(str(row_id), float(column))
                        columns[i]["cells"].append(dtofdp)

            if n % CHUNK_SIZE == 0:
                nchunks += 1
                consume_cells(columns)

        print(
            f"\r\t... reading from file is done, processed {nchunks} chunks of {CHUNK_SIZE} rows!"
        )

        consume_cells(columns)  # add remaining column values

        print("\r\t... data context is ready!")

    return data_context


# ---------------------------------------------------------------------------------
# MAIN


def main():
    folder_config = from_args()
    setup_logger_and_threads()

    print("Loading input data ...")
    input_data = read_csv_convert_to_aivis_format(
        os.path.join(folder_config["input_folder"], INPUT_DATA_FILE)
    )

    # print some record to illustrate type DtoFloatCell
    print("Example record!")
    some_column_id = "credit_amount"
    print("\t... Column id:", some_column_id)
    some_column_cells = input_data.get_float_column(some_column_id)
    print("\t... Column length:", len(some_column_cells))
    print("\t... Column entry number 42:", some_column_cells[42])
    print(
        "\t... Column entry number 42 row_id:",
        some_column_cells[42].id,
    )
    print("\t... Column entry number 42 value:", some_column_cells[42].value)

    # ---- ANALYZING ----

    print("Creating analysis ...")
    # build analysis config - for more options see documentation
    analysis_config = json.dumps(
        {
            "dataFilter": {"excludeColumns": ["telephone"]},
            "target": {
                "kpi": TARGET_COLUMN_ID,
                "interest": "HIGH_KPI",
            },
            "strategy": {
                "minimalFractionOfRecords": 0.05,
            },
        }
    )

    # create analysis
    analysis = ResponseAnalysis.create(input_data, analysis_config)
    print("\t... done!")

    print("Save analysis report ...")
    report = analysis.get_report()
    write_json(
        os.path.join(folder_config["output_folder"], "analysis-report.json"),
        json.loads(report),
    )

    print("\t... done!")

    # ---- Toolbox HTML Report ----
    from aivis_engine_v2_toolbox.api import build_response_analysis_report
    build_response_analysis_report(
        {"title": "Aivis Engine V2 - Response Analysis Example: German Credit Health",
         "archive": {"files": [os.path.join(folder_config["input_folder"], INPUT_DATA_FILE)]},
         "analysisReportFile": os.path.join(folder_config["output_folder"], "analysis-report.json"),
         "outputFile": os.path.join(folder_config["output_folder"], "aivis-engine-v2-response-analysis-example.html")})
    print("ALL DONE!")

# run main function
main()
